@@ -349,33 +349,4 @@ sub Get_DB_Handle {
 	return $code;
 }
 
-
 1;
-__END__
-
-=head1 NAME
-
-CIPP::DB_DBI - CIPP database module to generate Sybperl code
-
-=head1 DESCRIPTION
-
-CIPP has a database code abstraction layer, so it can
-generate code to access databases via different interfaces.
-
-This module is used by CIPP to generate code to access
-databases via Sybperl.
-
-=head1 AUTHOR
-
-Jrn Reder, joern@dimedis.de
-
-=head1 COPYRIGHT
-
-Copyright 1997-1999 dimedis GmbH, All Rights Reserved
-
-This library is free software; you can redistribute it and/or modify
-it under the same terms as Perl itself.
-
-=head1 SEE ALSO
-
-perl(1), CIPP (3pm)
